const {
  Mixin,
  set: set,
  get: get,
  getWithDefault,
  computed,
  isEmpty,
  on,
  Logger: { log, warn },
  run
} = Em

export default Mixin.create({
  attributeBindings: ['tabindex', 'role'],

  tabindex: 0,
  role: 'listbox',

  canRemoveSelection: true,
  optionLabelPath: 'label',
  optionValuePath: 'id',

  isOpen: false,
  isDisabled: false,
  isClosing: false,
  isFixed: false,

  options: Em.A([]),
  focused: null,

  search: null,
  disableSearch: false,

  label: null,

  selections: null,
  multiple: false,

  selectionComponent: computed('multiple', {
    get () {
      if (get(this, 'multiple')) {
        return 'en-multi-select-selection'
      } else {
        return 'en-select-selection'
      }
    }
  }),

  /**
   * hasSelections
   * 
   * Returns true if there are any elements
   * in the selections array
   *
   * @return {Boolean}
   */
  hasSelections: computed.gt('selections.length', 0),

  /**
   *  Selection
   *
   *  Returns the first object from selections unless
   *  in multiple mode.
   *
   * @return {Em.Object}
   */

  selection: computed('selections.[]', function () {
    const multiple   = get(this, 'multiple')
    const selections = get(this, 'selections')

    if (multiple) {
      warn('[en-select] Cannot get selection in multiple mode.')
      return
    }

    return get(selections, 'firstObject')
  }),

  hasSelection: computed.bool('selection'),

  /**
   * sortedOptions
   * 
   * By default, it's sorted by `id`. Can be overwrriten.
   *
   * @return {Em.Array}
   */

  sortProperties: ['id:asc'],
  sortedOptions: computed.sort('options', 'sortProperties'),

  /**
   * searchResults
   *
   * Computed property that filters the options if the
   * searched keyword is present in their `labelPath`.
   *
   * @return {Em.Array}
   */

  searchResults: computed('search', 'sortedOptions.[]', function () {
    const options   = get(this, 'sortedOptions')
    const search    = get(this, 'search')
    const labelPath = getWithDefault(this, 'optionLabelPath')
    
    let searchResults

    if (isEmpty(search)) {
      searchResults = options

    } else {
      searchResults = options.reject((option) => {
        const label = get(option, labelPath).toLowerCase()
        const keyword = search.toLowerCase()

        const strippedLabel = label.replace(/[^A-Za-z0-9]/g, '')
        const strippedKeyword = keyword.replace(/[^A-Za-z0-9]/g, '')

        return strippedLabel.indexOf(strippedKeyword) === -1
      })
    }

    return Em.A(searchResults)
  }),

  /**
   * @public
   * 
   * Selectable options; the difference between the
   * `searchResults` array and `selections` array.
   *
   * @return {Em.Array}
   */

  selectable: computed.setDiff('searchResults', 'selections'),

  /**
   * @private
   * 
   * handles clicks and sends the `onClose`
   * action if the click is outside the
   * select
   *
   * @return {undefined}
   */

  _handleClick (e) {
    if (!get(this, 'isOpen')) return

    const target = $(e.target)
    const select = $('.en-select')
    const optionClass = 'en-select-option'

    const parentClasses = target.parents().map(function() { return this.className.split(" ") })

    const isOptionClick = target.hasClass(optionClass) || ($.inArray(optionClass, parentClasses) !== -1)
    const isOutsideClick = !select.is(target) && select.has(target).length === 0 && !isOptionClick

    const shouldClose = isOptionClick || isOutsideClick

    const isMultiple = get(this, 'multiple')
    const noSelectablesRemaining = get(this, 'selectable').length === 0

    /*
     * In case it's a multiple box, we close only
     * when no selectables are remaining or when
     * clicked outside
     */

    if (isMultiple && (noSelectablesRemaining || isOutsideClick)) {
      this.send('closeSelect')
      this._tearDown()
    }

    /*
     * In other cases, we can close just when it's
     * outside the select box or on an option.
     */

    if (!isMultiple && shouldClose) {
      this.send('closeSelect')
      this._tearDown()
    }
  },


  /**
   * @private
   *
   * Moves to an option with an index specified
   * in relation to the currently focused option
   *
   * @return {undefined}
   */

  _move (direction, e) {
    const allowed = Em.A([1, -1])

    if (!allowed.contains(direction)) {
      warn(`[en-select] Cannot move in the direction provided - ${direction}`)
      return
    }

    const up   = direction === -1
    const down = direction === 1

    const options      = Em.A(get(this, 'selectable'))
    const currentFocus = get(this, 'focused')
    const indexOfFocus = options.indexOf(currentFocus)

    const lastIsFocused  = indexOfFocus === (get(options, 'length') - 1)
    const firstIsFocused = indexOfFocus === 0
    const noneFocused    = isEmpty(currentFocus)

    let nextFocus

    /**
     In case the first object is focused, or no element is focused,
     moving up will result in focusing the last element of the options.
    
     Similarly, in case the last object is focused, or no element is focused,
     moving down will result in focusing the first element.
    
     We can take those two situations out straightaway.
    */
    
    if ((firstIsFocused || noneFocused) && up) {
      nextFocus = get(options, 'lastObject')

    } else if ((lastIsFocused || noneFocused) && down) {
      nextFocus = get(options, 'firstObject')

    } else {
      nextFocus = options.objectAt(indexOfFocus + direction)
    }

    if (isEmpty(nextFocus)) {
      warn('[en-select] Could not find the next option to focus on')
      return
    }

    this.send('focus', nextFocus, e)
  },

  /**
   * @private
   * 
   * Finds the option above the currently
   * focused option and focuses on that
   *
   * @return {undefined}
   */

  _moveUp (e) {
    this._move(-1, e)
  },

  /**
   * @private
   * 
   * Finds the option below the currently
   * focused option and focuses on that
   *
   * @return {undefined}
   */

  _moveDown (e) {
    if (this.get('isOpen')) {
      this._move(1, e)
    } else {
      this.set('isOpen', true)
      this._move(1, e)
    }
  },

  /**
   * @private
   *
   * Selects the currently focused option.
   * Usually triggered through a keypress.
   *
   * @return {undefined}
   */

  _selectFocusedOption (e) {
    const focused = get(this, 'focused')

    if (!isEmpty(focused)) {
      this.send('select', focused)

      // In case this isn't a multiselect,
      // we have our option, and the select
      // options can be closed now.
      if (!get(this, 'multiple')) this.send('closeSelect')
    }
  },
  
  /**
   * @private
   *
   * Handles several different keyevents
   * and performs the right action.
   *
   * @return {undefined}
   */

  _handleKeydown (e) {
    let originalEvent = e.originalEvent

    switch (e.which) {
      case 38:
        e.stopPropagation()
        this._moveUp(originalEvent);
        break;

      case 40:
        e.stopPropagation()
        this._moveDown(originalEvent);
        break;

      case 13:
        e.stopPropagation();
        e.preventDefault();
        if (get(this, 'isOpen')) {
          this._selectFocusedOption(originalEvent);
        } else {
          this.send('openSelect')
        }

        break;

      case 27:
        e.stopPropagation()
        this.send('closeSelect')
        break;
    }
  },

  
  /**
   * @private
   *
   * Sets up the `click` and `keydown` handlers
   *
   * @return {undefined}
   */

  _setupAllHandlers () {
    $(document).on('click', this._handleClick)

    if (get(this, 'isFixed')) {
      $(window).on('resize', this._setupDropdownPositioning)
      $(window).on('scroll', this._setupDropdownPositioning)
    }
  },

  _setupDropdownPositioning () {
    let dropdown = this.$('.en-select-options')
    let { top, left } = $(this.element).offset()

    if (!top || !left) {
      console.warn('Could not find offset')
      return
    }

    let scrollTop = $(window).scrollTop()
    let elOffset = parseInt(top, 10) - parseInt(scrollTop, 10)

    let labelHeight = this.$('.en-select-label').height()
    let labelWidth  = this.$('.en-select-label').width()

    let offsetTop = parseInt(elOffset, 10) + parseInt(labelHeight, 10) + 15

    this.set('dropdownStyle', `top: ${offsetTop}px; left: ${left}px; width: ${labelWidth}px;`)
  },

  keyDown (e) {
    this._handleKeydown(e)
  },

  /**
   * @private
   *
   * Sets up the component when it's inserted.
   *
   * @return {undefined}
   */

  _setup: on('didInsertElement', function () {
    this._handleClick              = this._handleClick.bind(this)
    this._setupDropdownPositioning = this._setupDropdownPositioning.bind(this)
    this._handleKeydown            = this._handleKeydown.bind(this)
  }),

  /**
   * @private
   *
   * Tears down the component when it's about to destroy.
   *
   * @return {undefined}
   */

  _tearDown: on('willDestroyElement', function () {
    set(this, 'search', null)
    set(this, 'focused', null)

    $(document).off('click', this._handleClick)
    $(window).off('resize', this._setupDropdownPositioning)
    $(window).off('scroll', this._setupDropdownPositioning)
  }),

  /**
   * @private
   *
   *
   * @return {undefined}
   */

  _focusOnSearch () {
    run.next(() => {
      const searchBox = this.$('input[type="text"]')
      if (searchBox !== undefined) searchBox.focus()
    })
  },

  /**
   * @private
   * Select
   * 
   * Depending on the mode, it pushes the given `value`
   * into the selections array.
   *
   * @return {undefined}
   */

  _select (value) {
    const multiple = get(this, 'multiple')
    const valuePath = get(this, 'optionValuePath')

    const options  = this.getAttr('options')
    if (isEmpty(options)) return

    if (isEmpty(value)) {
      return multiple ? null : set(this, 'selections', Em.A([]))

    } else {
      const isObject = typeof value === "object"
      const id = isObject ? get(value, valuePath) : value

      const filtered = options.filter(option => {
        return get(option, valuePath) === id
      })

      if (multiple) {
        if (get(this, 'selections') === null) {
          set(this, 'selections', Em.A(filtered))
        } else {
          get(this, 'selections').addObjects(filtered)
        }

      } else {
        set(this, 'selections', Em.A(filtered))
      }
    }
  },

  _selectMultiple (values) {
    const options = this.getAttr('options')
    if (isEmpty(options) || isEmpty(values)) return

    const valuePath = get(this, 'optionValuePath')

    const mappedValues = values.map(value => {
      return (typeof value === "object") ? get(value, valuePath) : value
    })

    const filtered = options.filter(option => {
      return mappedValues.indexOf(get(option, valuePath)) > -1
    })

    set(this, 'selections', filtered)
  },

  /**
   * @private
   * 
   * Removes elements from selections array
   *
   * @return {undefined}
   */

  _removeFromSelections (value) {
    if (isEmpty(value)) {
      return
    } else {
      get(this, 'selections').removeObject(value)
    }
  },

  /**
   * @private
   * Updates selections whenever new attrs are received
   *
   */

  _updateSelections () {
    const values   = this.getAttr('value')
    const multiple = get(this, 'multiple')

    if (multiple) {
      this._selectMultiple(values)

    } else {
      this._select(values)
    }
  },

  didInitAttrs () {
    this._updateSelections()
  },

  didReceiveAttrs () {
    this._updateSelections()
  },

  /**
   * Init
   * Sets `selections` to a _new_ empty array. 
   * Not doing this will mean the previous 
   * array will get updated all the time.
   *
   * It also checks whether the `selection` is
   * null. `init` runs after the component has
   * received props, etc. so if it's filtered
   * out and a selection is set, it's not
   * supposed to change that.
   *
   * @return {undefined}
   */

  _setupInitialProps: on('init', function () {
    if (get(this, 'selections') === null) {
      set(this, 'selections', Em.A([]))
    }
  }),

  actions: {
    openSelect () {
      run.next(() => {
        if (get(this, 'isDisabled')) return

        set(this, 'isOpen', true)
        this._setupAllHandlers()

        if (get(this, 'isOpen')) {
          if (get(this, 'isFixed')) {
            this._setupDropdownPositioning()
          }

          this._focusOnSearch()
        }
      })
    },

    closeSelect () {
      set(this, 'isClosing', true)

      run.next(() => {
        if (get(this, 'isDestroyed')) return

        set(this, 'isOpen', false)
        set(this, 'isClosing', false)
        this._tearDown()
      }, 150)
    },

    searchFor (keyword) {
      set(this, 'search', keyword)

      const focused = get(this, 'searchResults').get('firstObject')
      set(this, 'focused', focused)

      if (!get(this, 'isOpen')) this.send('openSelect')
    },

    /**
     * Focuses on a given option, and, if the event
     * is a keyboard event, it asks the scroller to
     * scroll to that option.
     */

    focus (option, event) {
      set(this, 'focused', option)

      const isKeyboard = event && (event instanceof KeyboardEvent)
      if (isKeyboard) this.send('_scrollIfFocusNotVisible')

      this.sendAction('on-focus', option)
    },

    select (option) {
      set(this, 'search', null)
      this._select(get(option, 'id'))
      this.sendAction('on-select', option)

      if (get(this, 'isFixed')) {
        this._setupDropdownPositioning()
      }

      if (!get(this, 'multiple')) return

      run.next(() => {
        this.$('.en-select-search input').focus()
      })
    },

    removeSelection (option = null) {
      const multiple = get(this, 'multiple')

      if (multiple) {
        this._removeFromSelections(option)

      } else {
        set(this, 'selections', Em.A([]))
      }

      this.sendAction('on-remove', option)
    },

    /**
     *  Scroll to the focused option.
     *
     *  Checks if the focused option is out of view and
     *  scrolls to it, if so.
     */

    _scrollIfFocusNotVisible () {
      const list = this.$('.en-select-options').get(0)
      const focused = this.$('.en-select-option.is-focused').get(0)

      if (!list || !focused) return

      /**
       * Calculate the focused element's distance from the options
       * list's start
       */

      let scrollTop = focused.offsetTop - list.offsetTop

      /**
       * Add its offsetHeight to calculate the offset from top including
       * the focused element
       */

      let scrollBottom = scrollTop + focused.offsetHeight

      /**
       * Now, if the focused item's offset from the top is more than the
       * list's height + the list's current scroll position from the top,
       * you want to scroll again.
       */
      
      if (scrollBottom > (list.offsetHeight + list.scrollTop - 100)) {
        list.scrollTop = scrollBottom - (list.offsetHeight - 100)
      } else if (scrollTop < list.scrollTop) {
        list.scrollTop = scrollTop
      }
    }
  }
})
